/*
 * This file is a part of EZC -- Easy templating in C++ library
 * and is distributed under the BSD 3-Clause licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/* 
 * Copyright (c) 2007-2018, Tomasz Sowa
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *    
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *    
 *  * Neither the name Tomasz Sowa nor the names of contributors to this
 *    project may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef headerfile_ezc_item
#define headerfile_ezc_item

#include <vector>
#include <string>



namespace Ezc
{



struct Item
{
	enum Type 
	{
		item_none, item_container, item_text, item_function, item_if,
		item_for, item_else, item_end, item_err, item_include, item_comment,
		item_def, item_filter, item_ezc, item_block, item_return
	};


	struct Function
	{
		bool is_function;					// if true then we have a function parameter, if false then we have a string parameter
		std::wstring name;					// a function or parameter name
		std::wstring postfix;				// a function can have an additional postfix in its name
											// e.g. [my_function:my_postfix]
		std::vector<Function*> parameters;	// if is_function is true then it is a function and can have 'parameters'
											// if is_function is empty then 'parameters' is empty too
		void * fun_cache; 					// only valid if is_function is true
		Item * item_block;
		void * base_obj;
		int method_index;
		int arg;							// used if name is numeric (if no then is equal -1)


		Function()
		{
			is_function  = false;
			fun_cache    = 0;
			item_block   = 0;
			base_obj     = 0;
			method_index = -1;
			arg          = -1;
		}

		~Function()
		{
			Clear();
		}

		Function(const Function & f)
		{
			operator=(f);
		}

		Function & operator=(const Function & f)
		{
			is_function = f.is_function;
			name        = f.name;
			fun_cache   = f.fun_cache;
			item_block  = f.item_block;
			base_obj = f.base_obj;
			method_index = f.method_index;
			arg         = f.arg;

			for(size_t i=0 ; i<f.parameters.size() ; ++i)
				parameters.push_back(new Function(*f.parameters[i]));

		return *this;
		}

		Function & AddNewParam()
		{
			parameters.push_back(new Function());

		return *parameters.back();
		}

		void Clear()
		{
			for(size_t i=0 ; i<parameters.size() ; ++i)
				delete parameters[i];

			parameters.clear();
			name.clear();
			postfix.clear();
			fun_cache    = 0;
			item_block   = 0;
			base_obj     = 0;
			method_index = -1;
			arg          = -1;
			is_function  = false;
		}

	};



	Type type;
	std::wstring text;						// used in: item_text
	std::wstring file_name;					// used in: item_include (as a file name)
	std::vector<Item*> item_tab;				// childs
	bool has_function;						// the [return] statement can be without a function
	Function function;

	Item();
	Item(const Item & i);
	Item & operator=(const Item & i);
	void CopyItemTable(const Item & i);
	~Item();

	Item * AddItem(const Item * porg = 0);
	Item * AddItem(const Item & porg);
	void   ClearItems();
	Type   LastItemType();
	void   DeleteLastItem();
	void   Clear();
	void   ClearCache();


private:

	void ClearCache(Item::Function & function);

};




} // namespace Ezc


#endif
